# -*- coding: utf8 -*-
""" 命令行入口的协程版本
"""
from gevent import monkey;monkey.patch_all()
from .run import run as run_
from .subfinder_gevent import SubFinderGevent as SubFinder


def run():
    run_(SubFinder)


if __name__ == '__main__':
    run()


# -*- coding: utf8 -*-
""" 命令行入口的多线程版本
"""
from .subfinder_thread import SubFinderThread as SubFinder
from .run import run as run_


def run():
    run_(SubFinder)


if __name__ == '__main__':
    run()


# -*- coding: utf8 -*-
""" 命令行入口
"""
from __future__ import unicode_literals, print_function
import sys
import time
import argparse
import six
from .subsearcher import get_subsearcher, get_all_subsearchers, BaseSubSearcher
from .subfinder import SubFinder
from . import __version__


def find_method(m):
    s = get_subsearcher(m)
    if s is None:
        raise argparse.ArgumentTypeError(
            'Cant found SubSearcher named {}'.format(m))
    return s


def method_msg():
    all_subsearchers = get_all_subsearchers()
    support_methods = ', '.join(all_subsearchers.keys())
    default_subsearcher = get_subsearcher('default')
    msg = '''what's methods you want to use to searching subtitles, defaults to {default}.
            support methods: {support_methods}
        '''.format(default=default_subsearcher.__name__, support_methods=support_methods)
    return msg


def epilog():
    all_subsearchers = get_all_subsearchers()
    msg = 'Languages & Exts\n\n'
    for name, subsearcher in all_subsearchers.items():
        msg += 'languages supported by {}: {}\n'.format(
            name, subsearcher.SUPPORT_LANGUAGES)
        msg += 'exts supported by {}: {}\n'.format(
            name, subsearcher.SUPPORT_EXTS)
        msg += '\n'
    return msg


def run(subfinder_class):
    parser = argparse.ArgumentParser(prog='subfinder',
                                     description='A general subfinder, support for custom subsearcher',
                                     epilog=epilog(),
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     )
    parser.add_argument(
        'path', help="the video's filename or the directory contains vedio files")
    parser.add_argument('-l', '--languages',
                        nargs='+',
                        help="what's languages of subtitle you want to find")
    parser.add_argument('-e', '--exts',
                        nargs='+',
                        help="what's formats of subtitle you want to find")
    parser.add_argument('-m', '--method',
                        nargs='+', type=find_method, default='default',
                        help=method_msg())
    parser.add_argument('-s', '--silence',
                        action='store_true', default=False,
                        help="don't print anything, default to False")
    parser.add_argument('-p', '--pause',
                        action='store_true', default=False,
                        help="pause script after subfinder done. this option is used in 'Context Menu on Windows' only")
    parser.add_argument('-v', '--version',
                        action='version', version='subfinder {v}'.format(v=__version__),
                        help="show subfinder's version")
    parser.add_argument('--debug',
                        action='store_true', default=False,
                        help="print debug infomation, default to False")

    args = parser.parse_args()
    
    # try to make `path` to unicode string in python2
    if six.PY2 and isinstance(args.path, six.binary_type):
        args.path = args.path.decode(sys.getfilesystemencoding())

    subfinder = subfinder_class(path=args.path,
                                languages=args.languages,
                                exts=args.exts,
                                subsearcher_class=args.method,
                                silence=args.silence,
                                debug=args.debug)
    subfinder.start()
    subfinder.done()

    if args.pause:
        time.sleep(5)


if __name__ == '__main__':
    run(SubFinder)



# -*- coding: utf8 -*-
from __future__ import unicode_literals
import os
import sys
import logging
import mimetypes
import traceback
import requests
from .subsearcher import get_subsearcher, exceptions


class Pool(object):
    """ 模拟线程池，实际上还是同步执行代码
    """
    def __init__(self, size):
        self.size = size

    def spawn(self, fn, *args, **kwargs):
        fn(*args, **kwargs)

    def join(self):
        return


class SubFinder(object):
    """ 字幕查找器
    """

    VIDEO_EXTS = ['.mkv', '.mp4', '.ts', '.avi', '.wmv']

    def __init__(self, path='./', languages=None, exts=None, subsearcher_class=None, **kwargs):
        self.set_path(path)
        self.languages = languages
        self.exts = exts

        # silence: dont print anything
        self.silence = kwargs.get('silence', False)
        # logger's output
        self.logger_output = kwargs.get('logger_output', sys.stdout)
        # debug
        self.debug = kwargs.get('debug', False)
        self._init_session()
        self._init_pool()
        self._init_logger()

        # _history: recoding downloading history
        self._history = {}

        if subsearcher_class is None:
            subsearcher_class = get_subsearcher('default')
        if not isinstance(subsearcher_class, list):
            subsearcher_class = [subsearcher_class]

        self.subsearcher = [sc(self) for sc in subsearcher_class]

    def _is_videofile(self, f):
        """ 判断 f 是否是视频文件
        """
        if os.path.isfile(f):
            types = mimetypes.guess_type(f)
            mtype = types[0]
            if (mtype and mtype.split('/')[0] == 'video') or (os.path.splitext(f)[1] in self.VIDEO_EXTS):
                return True
        return False

    def _filter_path(self, path):
        """ 筛选出 path 目录下所有的视频文件
        """
        if self._is_videofile(path):
            return [path, ]

        if os.path.isdir(path):
            result = []
            for root, dirs, files in os.walk(path):
                result.extend(filter(self._is_videofile, map(
                    lambda f: os.path.join(root, f), files)))
            return result
        else:
            return []

    def _init_session(self):
        """ 初始化 requests.Session
        """
        self.session = requests.Session()
        self.session.mount('http://', adapter=requests.adapters.HTTPAdapter(
            pool_connections=10,
            pool_maxsize=100))

    def _init_pool(self):
        self.pool = Pool(10)

    def _init_logger(self):
        log_level = logging.INFO
        if self.silence:
            log_level = logging.CRITICAL + 1
        if self.debug:
            log_level = logging.DEBUG
        self.logger = logging.getLogger('SubFinder')
        self.logger.handlers = []
        self.logger.setLevel(log_level)
        sh = logging.StreamHandler(stream=self.logger_output)
        sh.setLevel(log_level)
        formatter = logging.Formatter(
            '[%(asctime)s]-[%(levelname)s]: %(message)s', datefmt='%m/%d %H:%M:%S')
        sh.setFormatter(formatter)
        self.logger.addHandler(sh)

    def _download(self, videofile):
        """ 调用 SubSearcher 搜索并下载字幕
        """
        basename = os.path.basename(videofile)

        subinfos = []
        for subsearcher in self.subsearcher:
            self.logger.info(
                '{0}：开始使用 {1} 搜索字幕'.format(basename, subsearcher))
            try:
                subinfos = subsearcher.search_subs(
                    videofile, self.languages, self.exts)
            except Exception as e:
                err = str(e)
                if self.debug:
                    err = traceback.format_exc()
                self.logger.error(
                    '{}：搜索字幕发生错误： {}'.format(basename, err))
                continue
            if subinfos:
                break
        self.logger.info('{1}：找到 {0} 个字幕, 准备下载'.format(
            len(subinfos), basename))
        try:
            for subinfo in subinfos:
                downloaded = subinfo.get('downloaded', False)
                if downloaded:
                    if isinstance(subinfo['subname'], (list, tuple)):
                        self._history[videofile].extend(subinfo['subname'])
                    else:
                        self._history[videofile].append(subinfo['subname'])
                else:
                    link = subinfo.get('link')
                    subname = subinfo.get('subname')
                    subpath = os.path.join(os.path.dirname(videofile), subname)
                    res = self.session.get(link, stream=True)
                    with open(subpath, 'wb') as fp:
                        for chunk in res.iter_content(8192):
                            fp.write(chunk)
                    self._history[videofile].append(subpath)
        except Exception as e:
            self.logger.error(str(e))

    def set_path(self, path):
        path = os.path.abspath(path)
        self.path = path

    def start(self):
        """ SubFinder 入口，开始函数
        """
        self.logger.info('开始')
        videofiles = self._filter_path(self.path)
        l = len(videofiles)
        if l == 0:
            self.logger.info(
                '在 {} 下没有发现视频文件'.format(self.path))
            return
        else:
            self.logger.info('找到 {} 个视频文件'.format(l))

        for f in videofiles:
            self._history[f] = []
            self.pool.spawn(self._download, f)
        self.pool.join()
        self.logger.info('='*20 + '下载完成' + '='*20)
        for v, subs in self._history.items():
            basename = os.path.basename(v)
            self.logger.info(
                '{}: 下载 {} 个字幕'.format(basename, len(subs)))

    def done(self):
       pass



# -*- coding: utf-8 -*-
""" 解压压缩包，支持zip, rar
"""
import os
import sys
import six

class CompressedFile(object):
    """ a simple wrapper class for ZipFile and RarFile, it's only support read.
    """
    EXTS = ['zip', 'rar']

    def __init__(self, file):
        self.file = file
        self._file = None
        _, ext = os.path.splitext(file)
        if ext == '.zip':
            import zipfile
            self._file = zipfile.ZipFile(self.file, 'r')
        elif ext == '.rar':
            import rarfile
            if sys.platform == 'win32':
                # if os system is windows，try to use built-in unrar
                rarfile.UNRAR_TOOL = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'unrar.exe')
            self._file = rarfile.RarFile(self.file, 'r')
        else:
            raise ValueError('CompressedFile doesnt support "{}"'.format(ext))

    @staticmethod
    def decode_file_name(name):
        if six.PY3:
            try:
                name = name.encode('cp437')
            except UnicodeEncodeError as e:
                pass
        if isinstance(name, six.binary_type):
            try:
                name = name.decode('gbk')
            except UnicodeDecodeError as e:
                try:
                    name = name.decode('utf8')
                except UnicodeDecodeError as e:
                    pass
        return name

    @classmethod
    def is_compressed_file(cls, filename):
        _, ext = os.path.splitext(filename)
        ext = ext[1:]
        return ext in cls.EXTS

    def isdir(self, name):
        info = self._file.getinfo(name)
        try:
            return info.isdir()
        except:
            return name.endswith(os.path.sep)

    def namelist(self):
        return self._file.namelist()

    def extract(self, filename, dest):
        f = self._file.open(filename, 'r')
        with open(dest, 'wb') as fp:
            fp.write(f.read())

    def close(self):
        self._file.close()


# -*- coding: utf8 -*-
from __future__ import unicode_literals, print_function
import os
import re
import bs4
from subfinder.tools.compressed_file import CompressedFile
from .subsearcher import BaseSubSearcher


class ZimuzuSubSearcher(BaseSubSearcher):
    """ zimuzu 字幕搜索器(http://www.zimuzu.io/)
    """
    SUPPORT_LANGUAGES = ['zh_chs', 'zh_cht', 'en', 'zh_en']
    SUPPORT_EXTS = ['ass', 'srt']
    LANGUAGES_MAP = {
        '简体': 'zh_chs',
        '繁體': 'zh_cht',
        '英文': 'en',
        '中英': 'zh_en',
    }
    COMMON_LANGUAGES = ['英文', '简体', '繁体']

    API = 'http://www.zimuzu.io/search/index'
    SUBTITLE_DOWNLOAD_LINK = 'http://www.subku.net/dld/'

    _cache = {}
    shortname = 'zimuzu'

    @classmethod
    def _gen_subname(cls, videofile, language, ext, **kwargs):
        language = []
        orig_subname = kwargs.get('orig_name')
        try:
            for l in cls.COMMON_LANGUAGES:
                if orig_subname.find(l) >= 0:
                    language.append(l)
        except Exception:
            pass
        language = '&'.join(language)
        basename = os.path.basename(videofile)
        basename, _ = os.path.splitext(basename)
        _, ext = os.path.splitext(orig_subname)
        return '{basename}.{language}{ext}'.format(
            basename=basename,
            language=language,
            ext=ext)

    def _parse_search_result_html(self, doc):
        """
        解析搜索结果页面，返回字幕信息列表
        """
        result = []
        soup = bs4.BeautifulSoup(doc, 'lxml')
        search_item_div_list = soup.select(
            '.search-result > ul > li > .search-item')
        if not search_item_div_list:
            return []
        for item in search_item_div_list:
            subinfo = {
                'title': '',
                'link': '',
                'author': 'zimuzu',
                'exts': [],
                'languages': [],
                'rate': 0,
                'download_count': 0,
            }
            a = item.find('a')
            if not a:
                continue
            url = a.get('href')
            subinfo['link'] = self._join_url(self.API, url)
            zh_title = a.get_text()
            for k, v in self.LANGUAGES_MAP.items():
                if k in zh_title:
                    subinfo['languages'].append(v)
            font = item.find('font', class_="f4")
            en_title = font.string
            subinfo['title'] = en_title
            result.append(subinfo)
        return result

    def _parse_detailpage_html(self, doc):
        """ 解析字幕详情页面
        """
        result = {
            'exts': [],
            'downloadpage_link': ''
        }
        soup = bs4.BeautifulSoup(doc, 'lxml')
        li_list = soup.select('ul.subtitle-info > li')
        for li in li_list:
            if li.string and '【格式】' in li.string:
                s = li.string.lower()
                for ext in self.SUPPORT_EXTS:
                    if ext in s:
                        result['exts'].append(ext)

        a = soup.select('.subtitle-links > a')
        if a:
            a = a[0]
            result['downloadpage_link'] = a.get('href')

        return result

    def _parse_downloadpage_html(self, doc):
        """ 解析下载页面，返回下载链接
        """
        soup = bs4.BeautifulSoup(doc, 'lxml')
        a = soup.select('.download-box > a.btn-click')
        if a:
            a = a[0]
            link = a.get('href')
            return link
        return ''

    def _first_filter_subinfo_list(self, subinfo_list, videoinfo, languages):
        season = videoinfo.get('season')
        episode = videoinfo.get('episode')

        result = []

        for subinfo in subinfo_list:
            title = subinfo.get('title')
            languages_ = subinfo.get('languages')
            videoinfo_ = self._parse_videoname(title)
            season_ = videoinfo_.get('season')
            episode_ = videoinfo_.get('episode')
            
            if (season == season_ and
                episode == episode_ and
                    set(languages_).intersection(set(languages))):

                result.append(subinfo)

        return result

    def _get_subinfo_list(self, keyword):
        """根据关键词搜索，返回字幕信息列表
        """
        res = self.session.get(
            self.API, params={'keyword': keyword, 'search_type': 'subtitle'})
        doc = res.content
        referer = res.url
        subinfo_list = self._parse_search_result_html(doc)

        return subinfo_list, referer

    def _visit_detailpage(self, detailpage_link, referer):
        """访问字幕详情页面，解析出下载页面的地址
        """
        headers = {'Referer': referer}
        res = self.session.get(detailpage_link, headers=headers)
        referer = res.url
        doc = res.content
        result = self._parse_detailpage_html(doc)
        return result, referer

    def _visit_downloadpage(self, downloadpage_link, referer):
        headers = {'Referer': referer}
        res = self.session.get(downloadpage_link, headers=headers)
        referer = res.url
        doc = res.content
        download_link = self._parse_downloadpage_html(doc)
        return download_link, referer

    def _get_keyword(self, videoinfo):
        """ 获取关键词
        """
        keyword = videoinfo.get('title')
        if videoinfo['season'] != 0:
            keyword += '.S{:02d}'.format(videoinfo['season'])
        if videoinfo['episode'] != 0:
            keyword += '.E{:02d}'.format(videoinfo['episode'])
        # replace space with '+'
        keyword = re.sub(r'\s+', '+', keyword)
        return keyword

    def _search_subs(self, videofile, languages, exts):
        videoname = self._get_videoname(videofile)
        videoinfo = self._parse_videoname(videoname)
        keyword = self._get_keyword(videoinfo)

        self._debug('videoinfo: {}'.format(videoinfo))

        # try find subinfo_list from self._cache
        if keyword not in self._cache:
            subinfo_list, referer = self._get_subinfo_list(keyword)
            self._cache[keyword] = (subinfo_list, referer)
        else:
            subinfo_list, referer = self._cache.get(keyword)

        self._debug('subinfo_list: {}'.format(subinfo_list))

        # 初步过滤掉无关的字幕
        subinfo_list = self._first_filter_subinfo_list(
            subinfo_list, videoinfo, languages)

        self._debug('subinfo_list: {}'.format(subinfo_list))

        # 补全字幕信息中的 exts 字段
        for subinfo in subinfo_list:
            detail_info, referer = self._visit_detailpage(
                subinfo['link'], referer)
            subinfo['exts'] = detail_info['exts']

        subinfo = self._filter_subinfo_list(
            subinfo_list, videoinfo, languages, exts)
        
        self._debug('subinfo: {}'.format(subinfo))

        if not subinfo:
            return []

        detail_info, referer = self._visit_detailpage(subinfo['link'], referer)
        downloadpage_link = detail_info['downloadpage_link']

        download_link, referer = self._visit_downloadpage(
            downloadpage_link, referer)

        filepath, referer = self._download_subs(
            download_link, videofile, referer, subinfo['title'])

        self._debug('filepath: {}'.format(filepath))

        subs = self._extract(filepath, videofile, exts)

        self._debug('subs: {}'.format(subs))

        return [{
            'link': referer,
            'language': subinfo['languages'],
            'ext': subinfo['exts'],
            'subname': subs,
            'downloaded': True
        }]



# -*- coding: utf8 -*-
""" SubFinder的多线程版本
"""

from __future__ import unicode_literals
from .subfinder import SubFinder
try:
    from queue import Queue
except ImportError as e:
    from Queue import Queue
from threading import Thread, Lock

class Pool(object):
    """ 线程池
    """
    def __init__(self, size):
        self.size = size
        self.queue = Queue(maxsize=size)
        self.threads = [Thread(target=self._run) for i in range(size)]
        self._lock = Lock()
        self.start_threads()

    def _acquire(self):
        self._lock.acquire()

    def _release(self):
        self._lock.release()
    
    def start_threads(self):
        for t in self.threads:
            t.daemon = True
            t.start()

    def _run(self):
        while True:
            fn, args, kwargs = self.queue.get()
            fn(*args, **kwargs)
            self.queue.task_done()

    def spawn(self, fn, *args, **kwargs):
        self.queue.put((fn, args, kwargs))
        
    def join(self):
        self.queue.join()

class SubFinderThread(SubFinder):
    """ SubFinder Thread version
    """
    def _init_pool(self):
        self.pool = Pool(10)




# -*- coding: utf8 -*-
""" SunFinder 的协程版本
"""

from __future__ import unicode_literals
from gevent.pool import Pool
from .subfinder import SubFinder

class SubFinderGevent(SubFinder):
    """ SubFinder Thread version
    """
    def _init_pool(self):
        self.pool = Pool(10)




# -*- coding: utf8 -*-

from .subsearcher import (
    BaseSubSearcher, 
    register,
    register_subsearcher, 
    get_subsearcher, 
    get_all_subsearchers,
)
from .shooter import ShooterSubSearcher
from .zimuku import ZimukuSubSearcher
from .zimuzu import ZimuzuSubSearcher

def _register_internal_subsearcher():
    register_subsearcher('default', ShooterSubSearcher)
    register_subsearcher(ShooterSubSearcher.shortname, ShooterSubSearcher)
    register_subsearcher(ZimukuSubSearcher.shortname, ZimukuSubSearcher)
    register_subsearcher(ZimuzuSubSearcher.shortname, ZimuzuSubSearcher)

_register_internal_subsearcher()


class InvalidFileError(Exception):
    pass


class LanguageError(Exception):
    pass


class ExtError(Exception):
    pass


class ShooterAPIError(Exception):
    pass


class ZimukuSearchingError(Exception):
    pass


from __future__ import unicode_literals
import os
import hashlib
import requests
from .subsearcher import BaseSubSearcher
from . import exceptions

class ShooterSubSearcher(BaseSubSearcher):
    """ find subtitles from shooter.org
    API URL: https://www.shooter.cn/api/subapi.php
    """
    shortname = 'shooter'
    API_URL = 'https://www.shooter.cn/api/subapi.php'
    SUPPORT_LANGUAGES = ['zh', 'en']
    SUPPORT_EXTS = ['ass', 'srt']

    SHOOTER_LANGUAGES_MAP = {
        'zh': 'Chn',
        'en': 'Eng'
    }

    def _search_subs(self, videofile, languages, exts):
        filehash = self._compute_video_hash(videofile)
        root, basename = os.path.split(videofile)
        payload = {'filehash': filehash,
                   'pathinfo': basename,
                   'format': 'json',
                   'lang': ''}

        result = {}
        for language in languages:
            payload['lang'] = self.SHOOTER_LANGUAGES_MAP.get(language)
            res = self.session.post(self.API_URL, data=payload)
            if res.status_code == requests.codes.ok:
                try:
                    result[language] = res.json()
                except Exception as e:
                    result[language] = []

        subinfos = []
        for language, subinfolist in result.items():
            ext_set = set()
            for subinfo in subinfolist:
                desc = subinfo['Desc']
                delay = subinfo['Delay']
                files = subinfo['Files']
                for item in files:
                    ext_ = item['Ext']
                    ext_ = ext_.lower()
                    link = item['Link']
                    if ext_ in exts and ext_ not in ext_set:
                        subinfos.append({
                            'link': link,
                            'language': language,
                            'subname': self._gen_subname(videofile, language, ext_),
                            'ext': ext_,
                            'downloaded': False
                        })
                        ext_set.add(ext_)
        return subinfos

    @staticmethod
    def _gen_subname(videofile, language, ext):
        """ generate filename of subtitles
        :TODO: fix the conflict of subname
        """
        root, basename = os.path.split(videofile)
        name, _ = os.path.splitext(basename)
        subname = '{basename}.{language}.{ext}'.format(
            basename=name,
            language=language,
            ext=ext)
        return subname

    @staticmethod
    def _compute_video_hash(videofile):
        """ compute videofile's hash
        reference: https://docs.google.com/document/d/1w5MCBO61rKQ6hI5m9laJLWse__yTYdRugpVyz4RzrmM/preview
        """
        seek_positions = [None] * 4
        hash_result = []
        with open(videofile, 'rb') as fp:
            total_size = os.fstat(fp.fileno()).st_size
            if total_size < 8192 + 4096:
                raise exceptions.InvalidFileError(
                    'the video[{}] is too small'.format(os.path.basename(videofile)))

            seek_positions[0] = 4096
            seek_positions[1] = total_size // 3 * 2
            seek_positions[2] = total_size // 3
            seek_positions[3] = total_size - 8192
            for pos in seek_positions:
                fp.seek(pos, 0)
                data = fp.read(4096)
                m = hashlib.md5(data)
                hash_result.append(m.hexdigest())
        return ';'.join(hash_result)


# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
import os
import sys
import glob
import argparse
import re
from .subsearcher import BaseSubSearcher


def rm_subtitles(path):
    """ delete all subtitles in path recursively
    """
    sub_exts = ['ass', 'srt', 'sub']
    count = 0
    for root, dirs, files in os.walk(path):
        for f in files:
            _, ext = os.path.splitext(f)
            ext = ext[1:]
            if ext in sub_exts:
                p = os.path.join(root, f)
                count += 1
                print('Delete {}'.format(p))
                os.remove(p)
    return count


def mv_videos(path):
    """ move videos in sub-directory of path to path.
    """
    count = 0
    for f in os.listdir(path):
        f = os.path.join(path, f)
        if os.path.isdir(f):
            for sf in os.listdir(f):
                sf = os.path.join(f, sf)
                if os.path.isfile(sf):
                    new_name = os.path.join(path, os.path.basename(sf))
                    try:
                        os.rename(sf, new_name)
                    except (WindowsError, OSError) as e:
                        print('mv {} happens error: {}'.format(sf, e))
                    else:
                        count += 1
                        print('mv {} to {}'.format(sf, new_name))
    return count


def rename_subtitle(source, template):
    """重命名字幕文件
    @param source: 要重命名的字幕文件
    @param template: 模板，例如：'Friends.S{season:02d}.E{episode:02d}.1080p.5.1Ch.BluRay.ReEnc-DeeJayAhmed.{}.{ext}'
    支持的变量有：season, episode, language, ext
    """
    root = os.path.dirname(source)
    old_name = os.path.basename(source)
    old_name, ext = os.path.splitext(old_name)
    ext = ext[1:]
    info = BaseSubSearcher._parse_videoname(old_name)
    s = info['season']
    e = info['episode']

    _, language = os.path.splitext(old_name)
    language = language[1:]
    if language == '':
        language = 'UNKNOWN'

    new_name = template.format(season=s, episode=e, language=language, ext=ext)
    new_file = os.path.join(root, new_name)
    os.rename(source, new_file)
    return new_file


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--delete-subs',
                        help="delete all subtitles in PATH recursively.")
    parser.add_argument('-m', '--move-videos',
                        help="move all files(like videos, subtitles) in sub-directory of PATH to PATH.")
    parser.add_argument('-r', '--rename-sub',
                        nargs=2,
                        help="rename subtitle.")

    args = parser.parse_args()

    if args.delete_subs:
        # try to decode str to unicode in python2
        try:
            args.delete_subs = args.delete_subs.decode(
                sys.getfilesystemencoding())
        except AttributeError as e:
            pass
        print('Start delete subtitles in {}'.format(args.delete_subs))
        c = rm_subtitles(args.delete_subs)
        print('Done, delete {} subtitles'.format(c))

    if args.move_videos:
        # try to decode str to unicode in python2
        try:
            args.move_videos = args.move_videos.decode(
                sys.getfilesystemencoding())
        except AttributeError as e:
            print('error')
            pass
        print('Start move videos in {}'.format(args.move_videos))
        c = mv_videos(args.move_videos)
        print('Done, move {} files'.format(c))

    if args.rename_sub:
        # try to decode str to unicode in python2
        source, template = args.rename_sub
        try:
            source = source.decode(sys.getfilesystemencoding())
        except AttributeError as e:
            print('error')
            pass
        source_list = glob.glob(source)
        try:
            template = template.decode(sys.getfilesystemencoding())
        except AttributeError as e:
            print('error')
            pass
        for source in source_list:
            new_file = rename_subtitle(source, template)
            print('rename {} -> {}'.format(source, new_file))


if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-
from __future__ import unicode_literals
from abc import abstractmethod, ABCMeta
import os
import re
import cgi
try:
    import urlparse
except ImportError as e:
    from urllib import parse as urlparse
import requests
from subfinder.tools.compressed_file import CompressedFile
from . import exceptions


registered_subsearchers = {}


def register_subsearcher(name, subsearcher_cls):
    """ register a subsearcher, the `name` is a key used for searching subsearchers.
    if the subsearcher named `name` already exists, then it's will overrite the old subsearcher.
    """
    if not issubclass(subsearcher_cls, BaseSubSearcher):
        raise ValueError(
            '{} is not a subclass of BaseSubSearcher'.format(subsearcher_cls))
    registered_subsearchers[name] = subsearcher_cls


def register(subsearcher_cls=None, name=None):
    def decorator(subsearcher_cls):
        if name is None:
            _name = subsearcher_cls.__name__
        else:
            _name = name
        register_subsearcher(_name, subsearcher_cls)
        return subsearcher_cls
    return decorator(subsearcher_cls) if subsearcher_cls is not None else decorator


def get_subsearcher(name, default=None):
    return registered_subsearchers.get(name, default)


def get_all_subsearchers():
    return registered_subsearchers


class BaseSubSearcher(object):
    """ The abstract class for search subtitles.

    You must implement following methods:
    - search_subs
    """
    __metaclass__ = ABCMeta

    SUPPORT_LANGUAGES = []
    SUPPORT_EXTS = []

    def __init__(self, subfinder,  **kwargs):
        """
        subfinder: SubFinder
        debug: 是否输出调试信息
        """
        self.session = requests.session()
        self.session.headers['User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/66.0.3359.181 Safari/537.36'
        self.subfinder = subfinder

    def _debug(self, msg):
        self.subfinder.logger.debug(msg)

    @classmethod
    def _check_languages(cls, languages):
        for lang in languages:
            if lang not in cls.SUPPORT_LANGUAGES:
                raise exceptions.LanguageError(
                    '{} don\'t support "{}" language'.format(cls.__name__, lang))

    @classmethod
    def _check_exts(cls, exts):
        for ext in exts:
            if ext not in cls.SUPPORT_EXTS:
                raise exceptions.ExtError(
                    '{} don\'t support "{}" ext'.format(cls.__name__, ext))

    @classmethod
    def _join_url(cls, url, path):
        """ join absolute `url` and `path`(href)
        """
        return urlparse.urljoin(url, path)

    @classmethod
    def _get_videoname(cls, videofile):
        """parse the `videofile` and return it's basename
        """
        name = os.path.basename(videofile)
        name = os.path.splitext(name)[0]
        return name

    RE_SEASON_EPISODE = re.compile(
        r'[Ss](?P<season>\d+)\.?[Ee](?P<episode>\d+)')
    RE_RESOLUTION = re.compile(r'(?P<resolution>720[Pp]|1080[Pp]|HR)')
    RE_SOURCE = re.compile(
        r'\.(?P<source>BD|BluRay|BDrip|WEB-DL|HDrip|HDTVrip|HDTV|HD|DVDrip)\.')
    RE_AUDIO_ENC = re.compile(
        r'(?P<audio_encoding>mp3|DD5\.1|DDP5\.1|AC3\.5\.1)')
    RE_VIDEO_ENC = re.compile(r'(?P<video_encoding>x264|H.264|AVC1|H.265)')

    @classmethod
    def _parse_videoname(cls, videoname):
        """ parse videoname and return video info dict
        video info contains:
        - title, the name of video
        - sub_title, the sub_title of video
        - resolution,
        - source,
        -
        - season, defaults to 0
        - episode, defaults to 0
        """
        info = {
            'title': '',
            'season': 0,
            'episode': 0,
            'sub_title': '',
            'resolution': '',
            'source': '',
            'audio_encoding': '',
            'video_encoding': '',
        }
        last_index = 0
        m = cls.RE_SEASON_EPISODE.search(videoname)
        if m:
            info['season'] = int(m.group('season'))
            info['episode'] = int(m.group('episode'))
            s, e = m.span()
            info['title'] = videoname[0:s].strip('.')
            last_index = e

        m = cls.RE_RESOLUTION.search(videoname)
        if m:
            info['resolution'] = m.group('resolution')
            s, e = m.span()
            if info['title'] == '':
                info['title'] = videoname[0:s].strip('.')

            if info['season'] > 0 and info['episode'] > 0:
                info['sub_title'] = videoname[last_index:s].strip('.')
            last_index = e

        if info['title'] == '':
            info['title'] = videoname

        m = cls.RE_SOURCE.search(videoname)
        if m:
            info['source'] = m.group('source')

        m = cls.RE_AUDIO_ENC.search(videoname)
        if m:
            info['audio_encoding'] = m.group('audio_encoding')

        m = cls.RE_VIDEO_ENC.search(videoname)
        if m:
            info['video_encoding'] = m.group('video_encoding')
        return info

    @classmethod
    def _gen_subname(cls, videofile, language, ext, **kwargs):
        """ 生成字幕文件名
        """
        root, basename = os.path.split(videofile)
        name, _ = os.path.splitext(basename)
        subname = '{basename}.{language}.{ext}'.format(
            basename=name,
            language=language,
            ext=ext)
        return subname

    @classmethod
    def _extract(cls, compressed_file, videofile, exts):
        """ 解压字幕文件，如果无法解压，则直接返回 compressed_file。
        exts 参数用于过滤掉非字幕文件，只有文件的扩展名在 exts 中，才解压该文件。
        """
        if not CompressedFile.is_compressed_file(compressed_file):
            return [compressed_file]

        root = os.path.dirname(compressed_file)
        subs = []
        cf = CompressedFile(compressed_file)
        for name in cf.namelist():
            if cf.isdir(name):
                continue
            # make `name` to unicode string
            orig_name = CompressedFile.decode_file_name(name)
            _, ext = os.path.splitext(orig_name)
            ext = ext[1:]
            if ext not in exts:
                continue
            subname = cls._gen_subname(videofile, '', '', orig_name=orig_name)
            subpath = os.path.join(root, subname)
            cf.extract(name, subpath)
            subs.append(subpath)
        cf.close()
        return subs

    @classmethod
    def _filter_subinfo_list(cls, subinfo_list, videoinfo, languages, exts):
        """ filter subinfo list base on:
        - season
        - episode
        - languages
        - exts
        -
        return a best matched subinfo
        """
        season = videoinfo.get('season')
        episode = videoinfo.get('episode')
        resolution = videoinfo.get('resolution')
        source = videoinfo.get('source')
        video_encoding = videoinfo.get('video_encoding')
        audio_encoding = videoinfo.get('audio_encoding')

        filtered_subinfo_list_1 = []
        filtered_subinfo_list_2 = []
        filtered_subinfo_list_3 = []
        filtered_subinfo_list_4 = []
        filtered_subinfo_list_5 = []
        filtered_subinfo_list = []

        for subinfo in subinfo_list:
            title = subinfo.get('title')
            videoinfo_ = cls._parse_videoname(title)
            season_ = videoinfo_.get('season')
            episode_ = videoinfo_.get('episode')
            resolution_ = videoinfo_.get('resolution')
            source_ = videoinfo_.get('source')
            video_encoding_ = videoinfo_.get('video_encoding')
            audio_encoding_ = videoinfo_.get('audio_encoding')
            languages_ = subinfo.get('languages')
            exts_ = subinfo.get('exts')

            if (season == season_ and
                episode == episode_ and
                set(languages_).intersection(set(languages)) and
                    set(exts_).intersection(set(exts))):

                filtered_subinfo_list_1.append(subinfo)
                if resolution_ == resolution:
                    filtered_subinfo_list_2.append(subinfo)
                    if source_ == source:
                        filtered_subinfo_list_3.append(subinfo)
                        if video_encoding_ == video_encoding:
                            filtered_subinfo_list_4.append(subinfo)
                            if audio_encoding_ == audio_encoding:
                                filtered_subinfo_list_5.append(subinfo)
        if filtered_subinfo_list_5:
            filtered_subinfo_list = filtered_subinfo_list_5
        elif filtered_subinfo_list_4:
            filtered_subinfo_list = filtered_subinfo_list_4
        elif filtered_subinfo_list_3:
            filtered_subinfo_list = filtered_subinfo_list_3
        elif filtered_subinfo_list_2:
            filtered_subinfo_list = filtered_subinfo_list_2
        elif filtered_subinfo_list_1:
            filtered_subinfo_list = filtered_subinfo_list_1

        if not filtered_subinfo_list:
            return None
        # sort by download_count and rate
        sorted_subinfo_list = sorted(filtered_subinfo_list,
                                     key=lambda item: (
                                         item['rate'], item['download_count']),
                                     reverse=True)
        return sorted_subinfo_list[0]

    def _download_subs(self, download_link, videofile, referer='', sub_title=''):
        """ 下载字幕
        videofile: 视频文件路径
        sub_title: 字幕标题（文件名）
        download_link: 下载链接
        referer: referer
        """
        root = os.path.dirname(videofile)
        name, _ = os.path.splitext(os.path.basename(videofile))
        ext = ''

        headers = {
            'Referer': referer
        }
        res = self.session.get(download_link, headers=headers, stream=True)
        referer = res.url

        # 尝试从 Content-Disposition 中获取文件后缀名
        content_disposition = res.headers.get('Content-Disposition', '')
        if content_disposition:
            _, params = cgi.parse_header(content_disposition)
            filename = params.get('filename')
            if filename:
                _, ext = os.path.splitext(filename)
                ext = ext[1:]

        if ext == '':
            # 尝试从url 中获取文件后缀名
            p = urlparse.urlparse(res.url)
            path = p.path
            if path:
                _, ext = os.path.splitext(path)
                ext = ext[1:]

        if ext == '':
            # 尝试从字幕标题中获取文件后缀名
            _, ext = os.path.splitext(sub_title)
            ext = ext[1:]

        filename = '{}.{}'.format(name, ext)
        filepath = os.path.join(root, filename)
        with open(filepath, 'wb') as fp:
            for chunk in res.iter_content(8192):
                fp.write(chunk)

        return filepath, referer

    def _search_subs(self, videofile, languages, exts):
        """ search subtitles of videofile.

        `videofile` is the absolute(or relative) path of the video file.

        `languages` is the language of subtitle, e.g chn, eng, the support for language is difference, depende on
        implemention of subclass. `languages` accepts one language or a list of language

        `exts` is the format of subtitle, e.g ass, srt, sub, idx, the support for ext is difference,
        depende on implemention of subclass. `ext` accepts one ext or a list of ext

        return a list of subtitle info
        [
            {
                'link': '',     # download link
                'language': '', # language
                'ext': '',      # ext
                'subname': '',  # the filename of subtitles
                'downloaded': False, # it's tell `SubFinder` whether need to download.
            },
            {
                'link': '',
                'language': '',
                'ext': '',
                'subname': '',
            },
            ...
        ]
        - `link`, it's optional, but if `downloaded` is False, then `link` is required.
        - `language`, it's optional
        - `subname`, it's optional, but if `downloaded` is False, then `subname` is required.
        - `downloaded`, `downloaded` is required.
            if `downloaded` is True, then `SubFinder` will not download again,
            otherwise `SubFinder` will download link.

        sub-class should implement this private method.
        """
        return []

    def search_subs(self, videofile, languages=None, exts=None):
        if languages is None:
            languages = self.SUPPORT_LANGUAGES
        elif isinstance(languages, str):
            languages = [languages]
        self._check_languages(languages)

        if exts is None:
            exts = self.SUPPORT_EXTS
        elif isinstance(exts, str):
            exts = [exts]
        self._check_exts(exts)

        return self._search_subs(videofile, languages, exts)

    def __str__(self):
        if hasattr(self.__class__, 'shortname'):
            name = self.__class__.shortname
        else:
            name = self.__class__.__name__
        return '<{}>'.format(name)

    def __unicode__(self):
        return self.__str__()



#!/usr/bin/env python
import re
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import six
from subfinder import __version__, __author__
version = __version__
author = __author__


if not version:
    raise RuntimeError('Cannot find version information')
if not author:
    raise RuntimeError('Cannot find author information')

requires = [
    'requests',
    'lxml',
    'bs4',
    'gevent',
    'rarfile',
    'six',
]


def readme():
    if six.PY2:
        with open('README.md') as fp:
            content = fp.read()
    else:
        with open('README.md', encoding='utf-8') as fp:
            content = fp.read()
    return content


setup(
    name='SubFinder',
    version=version,
    author=author,
    author_email='ljm51689@gmail.com',
    description='subfinder -- a general finder for subtitles',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/ausaki/subfinder/',
    keywords=['subtitle', 'subfinder', 'sub', 'subtitle finder', 'shooter', 'zimuku'],
    license='MIT',
    packages=['subfinder', 'subfinder.subsearcher', 'subfinder.tools'],
    include_package_data=True,
    python_requires='>=2.7',
    install_requires=requires,
    entry_points={
        'console_scripts': [
                'subfinder = subfinder.run_gevent:run',
                'subutils = subfinder.utils:main'
        ]
    },
)



# -*- coding: utf8 -*-
from __future__ import unicode_literals, print_function
import os
import re
import bs4
from .subsearcher import BaseSubSearcher


class ZimukuSubSearcher(BaseSubSearcher):
    """ zimuku 字幕搜索器(https://www.zimuku.cn/)
    """
    SUPPORT_LANGUAGES = ['zh_chs', 'zh_cht', 'en', 'zh_en']
    SUPPORT_EXTS = ['ass', 'srt']
    LANGUAGES_MAP = {
        '简体中文字幕': 'zh_chs',
        '简体中文': 'zh_chs',
        '繁體中文字幕': 'zh_cht',
        '繁體中文': 'zh_cht',
        'English字幕': 'en',
        'English': 'en',
        'english': 'en',
        '双语字幕': 'zh_en',
        '双语': 'zh_en'
    }
    COMMON_LANGUAGES = ['英文', '简体', '繁体']

    API = 'https://www.zimuku.cn/search'
    SUBTITLE_DOWNLOAD_LINK = 'http://www.subku.net/dld/'

    _cache = {}
    shortname = 'zimuku'

    @classmethod
    def _gen_subname(cls, videofile, language, ext, **kwargs):
        language = []
        orig_subname = kwargs.get('orig_name')
        try:
            for l in cls.COMMON_LANGUAGES:
                if orig_subname.find(l) >= 0:
                    language.append(l)
        except Exception:
            pass
        language = '&'.join(language)
        basename = os.path.basename(videofile)
        basename, _ = os.path.splitext(basename)
        _, ext = os.path.splitext(orig_subname)
        return '{basename}.{language}{ext}'.format(
            basename=basename,
            language=language,
            ext=ext)

    def _parse_downloadcount(self, text):
        """ parse download count
        text format maybe:
        - pure number: 1000
        - number + unit: 1万
        """
        unit_map = {
            '千': 1000,
            '万': 10000,
            '百万': 1000000,
        }

        m = re.match(r'^(\d+(?:\.\d+)?)(\w{0,2})$', text, re.UNICODE)
        if m:
            n = float(m.group(1))
            u = m.group(2)
            u = unit_map.get(u, 1)
            return int(n * u)
        else:
            return 0

    def _parse_search_results_html(self, doc):
        """ parse search result html, return subgroups
        subgroups: [{ 'title': title, 'link': link}]
        """
        subgroups = []
        soup = bs4.BeautifulSoup(doc, 'lxml')
        ele_divs = soup.select('div.item.prel')
        if not ele_divs:
            return subgroups
        for item in ele_divs:
            ele_a = item.select('p.tt > a')
            if not ele_a:
                continue
            link = ele_a[0].get('href')
            title = ele_a[0].get_text().strip()
            subgroups.append({
                'title': title,
                'link': link,
            })
        return subgroups

    def _parse_sublist_html(self, doc):
        soup = bs4.BeautifulSoup(doc, 'lxml')
        subinfo_list = []
        ele_tr_list = soup.select(
            'div.subs > table tr.odd, div.subs > table tr.even')
        if not ele_tr_list:
            return subinfo_list
        for tr in ele_tr_list:
            subinfo = {
                'title': '',
                'link': '',
                'author': '',
                'exts': [],
                'languages': [],
                'rate': 0,
                'download_count': 0,
            }
            ele_td = tr.find('td', class_='first')
            if ele_td:
                # 字幕标题
                subinfo['title'] = ele_td.a.get('title').strip()
                # 链接
                subinfo['link'] = ele_td.a.get('href').strip()
                # 格式
                ele_span_list = ele_td.select('span.label.label-info')
                for ele_span in ele_span_list:
                    ext = ele_span.get_text().strip()
                    ext = ext.lower()
                    ext = ext.split('/')
                    subinfo['exts'].extend(ext)
                # 作者
                ele_span = ele_td.select('span > a > span.label.label-danger')
                if ele_span:
                    subinfo['author'] = ele_span[0].get_text().strip()
            # 语言
            ele_imgs = tr.select('td.tac.lang > img')
            if ele_imgs:
                for ele_img in ele_imgs:
                    language = ele_img.get('title', ele_img.get('alt'))
                    language = self.LANGUAGES_MAP.get(language)
                    subinfo['languages'].append(language)
            # 评分
            ele_i = tr.select('td.tac i.rating-star')
            if ele_i:
                ele_i = ele_i[0]
                m = re.search(r'(\d+)', ele_i.get('title'))
                if m:
                    subinfo['rate'] = m.group(1)
            # 下载次数
            ele_td = tr.select('td.tac')
            if ele_td:
                ele_td = ele_td[-1]
                subinfo['download_count'] = self._parse_downloadcount(
                    ele_td.get_text().strip())
            subinfo_list.append(subinfo)
        return subinfo_list

    def _filter_subgroup(self, subgroups):
        """ choose a best subgroup from `subgroups`
        """
        subgroup = subgroups[0] if subgroups else None
        return subgroup

    def _get_subinfo_list(self, videoname):
        """ return subinfo_list of videoname
        """
        # searching subtitles
        res = self.session.get(self.API, params={'q': videoname})
        doc = res.content
        referer = res.url
        subgroups = self._parse_search_results_html(doc)
        if not subgroups:
            return []
        subgroup = self._filter_subgroup(subgroups)

        # get subtitles
        headers = {
            'Referer': referer
        }
        res = self.session.get(self._join_url(self.API, subgroup['link']))
        doc = res.content
        referer = res.url
        subinfo_list = self._parse_sublist_html(doc)
        return subinfo_list, referer

    def _visit_detailpage(self, detailpage_link, referer):
        detailpage_link = self._join_url(self.API, detailpage_link)
        m = re.search(r'(\d+)\.html', detailpage_link)
        if not m:
            return None
        l = '{}.html'.format(m.group(1))
        downloadpage_link = self._join_url(self.SUBTITLE_DOWNLOAD_LINK, l)
        return downloadpage_link, detailpage_link

    def _visit_downloadpage(self, downloadpage_link, referer):
        """ get the real download link of subtitles.
        """
        headers = {
            'Referer': referer
        }
        res = self.session.get(downloadpage_link, headers=headers)
        doc = res.content
        referer = res.url
        soup = bs4.BeautifulSoup(doc, 'lxml')
        ele_a_list = soup.select('a.btn.btn-sm')
        if not ele_a_list:
            return None
        ele_a = ele_a_list[1]
        download_link = ele_a.get('href')
        return download_link, referer

    def _search_subs(self, videofile, languages, exts):
        videoname = self._get_videoname(videofile)  # basename, not include ext
        videoinfo = self._parse_videoname(videoname)
        keyword = videoinfo.get('title')
        if videoinfo['season'] != 0:
            keyword += '.S{:02d}'.format(videoinfo['season'])

        self._debug('videoinfo: {}'.format(videoinfo))

         # try find subinfo_list from self._cache
        if keyword not in self._cache:
            subinfo_list, referer = self._get_subinfo_list(keyword)
            self._cache[keyword] = (subinfo_list, referer)
        else:
            subinfo_list, referer = self._cache.get(keyword)

        self._debug('subinfo_list: {}'.format(subinfo_list))

        subinfo = self._filter_subinfo_list(
            subinfo_list, videoinfo, languages, exts)

        self._debug('subinfo: {}'.format(subinfo))

        if not subinfo:
            return []
            
        downloadpage_link, referer = self._visit_detailpage(
            subinfo['link'], referer)

        subtitle_download_link, referer = self._visit_downloadpage(
            downloadpage_link, referer)

        filepath, referer = self._download_subs(
            subtitle_download_link, videofile, referer, subinfo['title'])

        self._debug('filepath: {}'.format(filepath))

        subs = self._extract(filepath, videofile, exts)

        self._debug('subs: {}'.format(subs))

        return [{
            'link': referer,
            'language': subinfo['languages'],
            'ext': subinfo['exts'],
            'subname': subs,
            'downloaded': True
        }]
